/*
 *  Copyright 2010 Jesse Long.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *  under the License.
 */

package net.rumati.sqlblocks.compile;

import java.util.List;
import net.rumati.sqlblocks.AndWhereConditionList;
import net.rumati.sqlblocks.BaseSQLCompiler;
import net.rumati.sqlblocks.ColumnExpression;
import net.rumati.sqlblocks.ColumnParameterInfo;
import net.rumati.sqlblocks.ComparativeWhereCondition;
import net.rumati.sqlblocks.DeleteCommand;
import net.rumati.sqlblocks.ExistsCondition;
import net.rumati.sqlblocks.Expression;
import net.rumati.sqlblocks.InsertCommand;
import net.rumati.sqlblocks.IsNotNullCondition;
import net.rumati.sqlblocks.IsNullCondition;
import net.rumati.sqlblocks.Join;
import net.rumati.sqlblocks.MathExpression;
import net.rumati.sqlblocks.NegatedWhereCondition;
import net.rumati.sqlblocks.NumericValue;
import net.rumati.sqlblocks.OrWhereConditionList;
import net.rumati.sqlblocks.OrderExpression;
import net.rumati.sqlblocks.OutputColumn;
import net.rumati.sqlblocks.ParameterExpression;
import net.rumati.sqlblocks.Relation;
import net.rumati.sqlblocks.Schema;
import net.rumati.sqlblocks.SelectQuery;
import net.rumati.sqlblocks.Table;
import net.rumati.sqlblocks.UpdateCommand;
import net.rumati.sqlblocks.WhereCondition;
import net.rumati.sqlblocks.WhereConditionList;

/**
 * Converts SQL commands built from SQLBlocks classes into strings containing SQL strings.
 * @author jesse
 */
public class SQLCompiler
    extends BaseSQLCompiler
{

    @Override
    protected void appendAndWhereConditionListSQL(AndWhereConditionList awcl, ResultBuilder result)
    {
        boolean first = true;
        for (WhereCondition iwc : awcl.getConditions()){
            if (iwc instanceof WhereConditionList && ((WhereConditionList)iwc).numberOfConditions() < 1){
                // ignore
            }else{
                if (first){
                    first = false;
                }else{
                    result.appendNewLine();
                    result.append("AND");
                    result.appendNewLine();
                }
                if (iwc instanceof OrWhereConditionList && ((OrWhereConditionList)iwc).numberOfConditions() > 1){
                    result.append('(');
                    result.appendNewLine();
                    result.indent();
                    appendWhereConditionSQL(iwc, result);
                    result.unindent();
                    result.appendNewLine();
                    result.append(')');
                }else{
                    appendWhereConditionSQL(iwc, result);
                }
            }
        }
    }

    protected void appendDeleteCommandSQL(DeleteCommand cmd, ResultBuilder r)
    {
        r.append("DELETE FROM ");
        appendTableName(cmd.getTable(), r);
        appendWherePhraseSQL(cmd.getWhereCondition(), r);
    }

    protected void appendGroupByPhraseSQL(SelectQuery q, ResultBuilder r)
    {
        List<Expression> gpc = q.getGroupByColumns();
        if (gpc.size() > 0){
            r.appendNewLine();
            r.append("GROUP BY");
            r.appendNewLine();
            r.indent();
            boolean first = true;
            for (Expression gc : gpc){
                if (first){
                    first = false;
                }else{
                    r.append(',');
                    r.appendNewLine();
                }
                appendExpressionSQL(gc, r);
            }
            r.unindent();
        }
    }

    protected void appendInsertCommandSQL(InsertCommand cmd, ResultBuilder r)
    {
        r.append("INSERT INTO ");
        appendTableName(cmd.getTable(), r);
        r.append(" (");
        boolean first = true;
        for (ColumnParameterInfo c : getCommandColumnParameterInfo(cmd)){
            if (first){
                first = false;
            }else{
                r.append(", ");
            }
            appendName(c.getColumnName(), r);
        }
        r.append(") VALUES (");
        first = true;
        for (ColumnParameterInfo c : getCommandColumnParameterInfo(cmd)){
            if (first){
                first = false;
            }else{
                r.append(", ");
            }
            appendExpressionSQL(new ParameterExpression(c.getParameterKey()), r);
        }
        r.append(')');
    }

    protected void appendSelectQuerySQL(SelectQuery query, ResultBuilder r)
    {
        r.append("SELECT");
        r.appendNewLine();
        r.indent();
        appendSelectQueryOutputColumns(query, r);
        r.unindent();

        appendSelectQueryRelationList(query, r);

        appendWherePhraseSQL(query.getWhereCondition(), r);

        appendGroupByPhraseSQL(query, r);

        appendOrderByPhraseSQL(query, r);

        if (query.getLimit() > 0){
            r.appendNewLine();
            r.append("LIMIT ");
            r.append("" + query.getLimit());
        }

        if (query.getOffset() > 0){
            r.appendNewLine();
            r.append("OFFSET ");
            r.append("" + query.getOffset());
        }
    }

    protected void appendRelationWithAlias(Relation rel, String alias, ResultBuilder r)
    {
        appendRelationDefinition(rel, r);
        r.append(" AS ");
        appendName(alias, r);
    }

    protected void appendSelectQueryRelationList(SelectQuery query, ResultBuilder r)
    {
        r.appendNewLine();
        r.append("FROM");
        r.appendNewLine();
        appendRelationWithAlias(query.getFrom(), query.getFromAlias(), r);

        for (Join j : query.getJoins()){
            r.appendNewLine();
            switch (j.getJoinType()){
                case INNER:
                    r.append("INNER JOIN ");
                    break;
                case LEFT_OUTER:
                    r.append("LEFT OUTER JOIN ");
                    break;
                case RIGHT_OUTER:
                    r.append("RIGHT OUTER JOIN ");
                    break;
            }
            appendRelationWithAlias(j.getRelation(), j.getJoinAlias(), r);
            WhereCondition joinCondition = j.getJoinCondition();
            if (joinCondition != null){
                r.append("ON");
                r.appendNewLine();
                r.indent();
                appendWhereConditionSQL(joinCondition, r);
                r.unindent();
            }
        }
    }

    protected void appendWherePhraseSQL(WhereCondition wc, ResultBuilder r)
    {
        if (wc != null && !(wc instanceof WhereConditionList && ((WhereConditionList)wc).numberOfConditions() < 1)){
            r.appendNewLine();
            r.append("WHERE");
            r.appendNewLine();
            r.indent();
            appendWhereConditionSQL(wc, r);
            r.unindent();
        }
    }

    protected void appendOrWhereConditionListSQL(OrWhereConditionList owc, ResultBuilder result)
    {
        boolean first = true;
        for (WhereCondition iwc : owc.getConditions()){
            if (iwc instanceof WhereConditionList && ((WhereConditionList)iwc).numberOfConditions() < 1){
                // ignore
            }else{
                if (first){
                    first = false;
                }else{
                    result.appendNewLine();
                    result.append("OR");
                    result.appendNewLine();
                }
                if (iwc instanceof AndWhereConditionList && ((AndWhereConditionList)iwc).numberOfConditions() > 1){
                    result.append('(');
                    result.appendNewLine();
                    result.indent();
                    appendWhereConditionSQL(iwc, result);
                    result.unindent();
                    result.appendNewLine();
                    result.append(')');
                }else{
                    appendWhereConditionSQL(iwc, result);
                }
            }
        }
    }

    @Override
    protected void appendComparativeWhereConditionSQL(ComparativeWhereCondition c, ResultBuilder r)
    {
        appendExpressionSQL(c.getLeftHandExpression(), r);
        r.append(' ');
        r.append(getComparisonOperatorSQL(c.getOperator()));
        r.append(' ');
        appendExpressionSQL(c.getLeftHandExpression(), r);
    }

    protected void appendColumnExpressionSQL(ColumnExpression col, ResultBuilder result)
    {
        appendName(col.getRelationAlias(), result);
        result.append(".");
        appendName(col.getColumnName(), result);
    }

    @Override
    protected void appendExistsConditionSQL(ExistsCondition c, ResultBuilder r)
    {
        r.append("EXISTS (");
        r.appendNewLine();
        r.indent();
        appendSelectQuerySQL(c.getSubQuery(), r);
        r.unindent();
        r.appendNewLine();
        r.append(")");
    }

    protected void appendIsNotNullWhereConditionSQL(IsNotNullCondition c, ResultBuilder result)
    {
        appendExpressionSQL(c.getColumn(), result);
        result.append(" IS NOT NULL");
    }

    protected void appendIsNullWhereConditionSQL(IsNullCondition c, ResultBuilder result)
    {
        appendExpressionSQL(c.getColumn(), result);
        result.append(" IS NULL");
    }

    protected void appendMathExpressionSQL(MathExpression ex, ResultBuilder result)
    {
        appendExpressionSQL(ex.getLeftHandExpression(), result);
        result.append(" ");
        result.append(getMathOperatorSQL(ex.getOperator()));
        result.append(" ");
        appendExpressionSQL(ex.getRightHandExpression(), result);
    }

    protected void appendName(String name, ResultBuilder r)
    {
        r.append('"');
        r.append(name.replace("\\", "\\\\").replace("\"", "\\\""));
        r.append('"');
    }

    protected void appendNegatedWhereConditionSQL(NegatedWhereCondition nc, ResultBuilder result)
    {
        WhereCondition iwc = nc.getNegatedCondition();
        result.append("NOT (");
        if (iwc instanceof AndWhereConditionList || iwc instanceof OrWhereConditionList){
            result.appendNewLine();
            result.indent();
            appendWhereConditionSQL(iwc, result);
            result.unindent();
            result.appendNewLine();
        }else{
            appendWhereConditionSQL(iwc, result);
        }
        result.append(')');
    }

    protected void appendNumericValueSQL(NumericValue v, ResultBuilder r)
    {
        r.append(v.getValue().toPlainString());
    }

    protected void appendOrderByPhraseSQL(SelectQuery query, ResultBuilder r)
    {
        List<OrderExpression> orderExpressions = query.getOrderExpressions();
        if (orderExpressions.size() > 0){
            r.appendNewLine();
            r.append("ORDER BY");
            r.appendNewLine();
            r.indent();
            boolean first = true;
            for (OrderExpression oe : orderExpressions){
                if (first){
                    first = false;
                }else{
                    r.append(',');
                    r.appendNewLine();
                }
                appendExpressionSQL(oe.getExpression(), r);
                switch (oe.getDirection()){
                    case ASCENDING:
                        r.append(" ASC");
                        break;
                    case DESCENDING:
                        r.append(" DESC");
                        break;
                }
            }
            r.unindent();
        }
    }

    protected void appendParameterExpressionSQL(ParameterExpression ex, ResultBuilder res)
    {
        res.append("?");
        res.addParameter(ex.getKey());
    }

    protected void appendSelectQueryOutputColumns(SelectQuery query, ResultBuilder r)
    {
        boolean first = true;
        List<OutputColumn> outputColumns = query.getOutputColumns();
        if (outputColumns.size() == 0){
            r.append('*');
        }else{
            for (OutputColumn oc : query.getOutputColumns()){
                if (first){
                    first = false;
                }else{
                    r.append(',');
                    r.appendNewLine();
                }
                appendExpressionSQL(oc.getExpression(), r);
                r.append(" AS ");
                appendName(oc.getAlias(), r);
            }
        }
    }

    @Override
    protected void appendSubQuerySQL(SelectQuery subQuery, ResultBuilder r)
    {
        r.append('(');
        r.appendNewLine();
        r.indent();
        appendSelectQuerySQL(subQuery, r);
        r.unindent();
        r.appendNewLine();
        r.append(")");
    }

    @Override
    protected void appendTableName(Table t, ResultBuilder r)
    {
        Schema s = t.getSchema();
        if (s != null){
            appendName(s.getName(), r);
            r.append('.');
        }
        appendName(t.getTableName(), r);
    }

    protected void appendUpdateCommandSQL(UpdateCommand cmd, ResultBuilder r)
    {
        r.append("UPDATE ");
        appendTableName(cmd.getTable(), r);
        boolean first = true;
        for (ColumnParameterInfo c : getCommandColumnParameterInfo(cmd)){
            if (first){
                r.append(" SET");
                r.appendNewLine();
                r.indent();
                first = false;
            }else{
                r.append(',');
                r.appendNewLine();
            }
            appendName(c.getColumnName(), r);
            r.append(" = ");
            appendExpressionSQL(new ParameterExpression(c.getParameterKey()), r);
        }
        if (!first){
            r.unindent();
        }
        appendWherePhraseSQL(cmd.getWhereCondition(), r);
    }

    public final Result compile(InsertCommand cmd)
    {
        ResultBuilder r = new ResultBuilder();
        appendInsertCommandSQL(cmd, r);
        return r.getResult();
    }

    public final Result compile(DeleteCommand cmd)
    {
        ResultBuilder r = new ResultBuilder();
        appendDeleteCommandSQL(cmd, r);
        return r.getResult();
    }

    public final Result compile(SelectQuery query)
    {
        ResultBuilder r = new ResultBuilder();
        appendSelectQuerySQL(query, r);
        return r.getResult();
    }

    public final Result compile(UpdateCommand cmd)
    {
        ResultBuilder r = new ResultBuilder();
        appendUpdateCommandSQL(cmd, r);
        return r.getResult();
    }
}
